from .abstract import ConfigObject
from ywh2bt.logging import logger
import importlib
import os
import sys
from ywh2bt.utils import read_input
from colorama import Fore, Style

__all__ = ["ExtraPackageConfig"]

"""
Module to load and configure External BugTracker packages to append.
"""


class ExtraPackageConfig(ConfigObject):
    """
    Load / Configure External packages to append.
    """

    ############################################################
    ###################### Constructor #########################
    ############################################################
    def __init__(self, configure_mode=False, **config):
        self._configure_mode = configure_mode
        self._package_name = ""
        keys = []
        self._modules = []
        self._path = ""
        if config or not configure_mode:
            keys += ["path", "modules"]
            self._path = config["path"]
            self._modules = [module for module in config["modules"]]
            if "package" in config:
                self._package_name = config["package"]

        super().__init__(keys, "Packages", **config)
        if configure_mode:
            self.configure()
        self.bugtracker_class_appender()

    ############################################################
    #################### Instance Methods ######################
    ############################################################
    def bugtracker_class_appender(self):
        """
        Verify if package exist and append it to sys.path.
        Import BugTrackerConfig module definition
        """
        if not os.path.isdir(self.format_path()):
            raise FileNotFoundError(
                "package path does'nt exist: {}".format(
                    os.path.abspath(self.path)
                )
            )

        sys.path.append(self.path)

        for bugtracker_module in self.modules:
            module_path = ""
            if self.package_name:
                module_path += self.package_name + "."
            module_path += bugtracker_module
            importlib.import_module(module_path)

    def configure(self):
        """
        Configure packages interactively.
        """
        exit_config = False
        self._package_name = read_input(
            Fore.BLUE
            + "Package Name (empty for module path): "
            + Style.RESET_ALL
        )
        while True:
            self._path = read_input(
                Fore.BLUE
                + "Absolute path to package directory: "
                + Style.RESET_ALL
            )
            if not os.path.isdir(self.format_path()):
                logger.error(
                    "package path does'nt exist: {}".format(self.path)
                )
                continue
            break

        while not exit_config:
            self._modules.append(
                read_input(Fore.BLUE + "Module name: " + Style.RESET_ALL)
            )
            while True:
                exit_configuration = read_input(
                    Fore.BLUE + "Add module ? [y/N]: " + Style.RESET_ALL
                )
                if exit_configuration in ["n", "N", ""]:
                    exit_config = True
                    break
                elif exit_configuration in ["y", "Y"]:
                    break
        self.info()

    def format_path(self):
        """
        Normalize given path.
        """
        return os.path.realpath(os.path.expanduser(self._path))

    def info(self):
        """
        Log information for the package.
        """
        logger.info("package name : {}".format(self.package_name))
        logger.info("path : {}".format(self.path))

        logger.info(
            "Modules : {}".format(
                "\n\t".join(
                    [
                        "{}/ {}".format(idx + 1, module)
                        for idx, module in enumerate(self.modules)
                    ]
                )
            )
        )

    def to_dict(self):
        """
        Map object to dictionary.
        """
        return {
            "package": self.package_name,
            "path": self.path,
            "modules": self.modules,
        }

    ############################################################
    ####################### Properties #########################
    ############################################################
    @property
    def package_name(self):
        return self._package_name

    @property
    def configure_mode(self):
        return self._configure_mode

    @property
    def path(self):
        return self._path

    @property
    def modules(self):
        return self._modules



# -*- encoding: utf-8 -*-
import io
import base64
import imghdr
import mimetypes
from abc import abstractmethod
import html2text
import urllib.parse as uparse
import re
from bs4 import BeautifulSoup

__all__ = ["BugTracker"]

"""
Define Abstract client class
"""


def html_handler(html):
    """
    Process html to markdown
    """
    html_parser = html2text.HTML2Text()
    html_parser.body_width = 0
    html_parser.mark_code = True
    langs = []
    soup = BeautifulSoup(html, "lxml")
    for pre in soup.findAll("pre"):
        for code in pre:
            langs.append(
                code.attrs.get("class", [""])[0].replace("language-", "")
            )
    n_html = html_parser.handle(html)
    idx = 0
    for i in range(len(langs)):
        idx = n_html[idx:].index("[code]") + 6 + idx
        n_html = n_html[:idx] + langs[i] + n_html[idx:]
    h = n_html.replace("[code]", "```").replace("[/code]", "```")
    return h


class BugTracker(object):

    """
    Abstract client Wrapper class for all BugTrackers Client.

    :attr str issue_name_template: template for the title part.
    :attr str ywh_comment_marker: template for comment marker part.
    :attr str ywh_comment_marker: template for comment part.
    :attr str description_template: template for description part.
    """

    issue_name_template = "{local_id} : {title}"
    ywh_comment_marker = (
        "Imported to the bugtracker : {url} on project : {project_id}."
    )
    ywh_comment_template = "Tracked to [{type} #{issue_id}]({bug_url})"

    description_template = """
| Title | {local_id} : {title}|
|-------|---------------------|
| Priority | {priority__name} |
| Bug Type | [{bug_type__name}]({bug_type__link}) &#8594; [Remediation]({bug_type__remediation_link}) |
| Scope | {scope} |
| Severity | {cvss__criticity}, score: {cvss__score:.1f}, vector: {cvss__vector} |
| Endpoint | {end_point} |
| Vulnerable part | {vulnerable_part} |
| Part Name | {part_name} |
| Payload | {payload_sample} |
| Technical Environment | {technical_information} |

{description_html}
    """

    ############################################################
    #################### Instance methods ######################
    ############################################################
    def file_to_base64(self, img, content_type):
        """
        Transforme input image to base64 string.

        :param bytes img: img to change to base64?
        :param content_type: content_type header
        """
        f = io.BytesIO(img)
        mime = imghdr.what(f)
        if mime is None and (
            not content_type
            or "image/" not in content_type
            or "svg" in content_type
        ):
            raise ValueError("Image type not supported")
        mime = mimetypes.types_map["." + mime] if mime else content_type
        image = (
            "data:"
            + mime
            + ";base64,"
            + str(base64.b64encode(img).decode("utf-8"))
        )
        return image

    def replace_external_link(self, report):
        description = report.description_html
        base_string = "https://{ywhdomain}/redirect?url=".format(
            ywhdomain=self._ywh_domain
        )

        pattern = re.compile(
            '"{base_string}([^ "]*)"'.format(
                base_string=base_string.replace("/", "\/")
                .replace("?", "\?")
                .replace(".", "\.")
            )
        )
        redirect_urls = pattern.findall(report.description_html)
        for url in redirect_urls:
            base_url, params = uparse.splitquery(uparse.unquote(url))
            new_params = "&".join(
                [
                    p
                    for p in params.split("&")
                    if not p.startswith("expires")
                    and not p.startswith("token")
                ]
            )
            description = description.replace(
                "{base_string}{url}".format(base_string=base_string, url=url),
                "{base_url}{params}".format(
                    base_url=base_url
                    + ("/" if not base_url.endswith("/") else ""),
                    params=("?" + new_params) if new_params else "",
                ),
            )
        return description

    def report_as_title(
        self,
        report,
        template=None,
        additional_keys=[],
        html_parser=html2text.html2text,
    ):
        """
        Format template title with given information.

        :param yeswehack.api.Report report: Report representation
        :param str template: if not None, overwride default template.
        :param list additional_keys: keys corresponding to attribute of report object.

        to used subjobject, use '__' separator : cvss__score -> cvss.score
        """
        if template is None:
            template = self.issue_name_template
        return self.format_template(
            report,
            template,
            keys=additional_keys,
            fmt_keys={"local_id": report.local_id, "title": report.title},
            html_parser=html_parser,
        )

    def report_as_description(
        self,
        report,
        template=None,
        additional_keys=[],
        html_parser=html_handler,
    ):
        """
        Format template description with given information

        :param yeswehack.api.Report report: Report representation
        :param str template: if not None, overwride default template.
        :param list additional_keys: keys corresponding to attribute of report object.

        to used subjobject, use '__' separator : cvss__score -> cvss.score
        """
        report.description_html = self.replace_external_link(report)
        if template is None:
            template = self.description_template
        keys = [
            "scope",
            "cvss__score",
            "priority__name",
            "cvss__criticity",
            "cvss__vector",
            "vulnerable_part",
            "bug_type__name",
            "bug_type__link",
            "bug_type__remediation_link",
            "description_html",
            "end_point",
            "vulnerable_part",
            "part_name",
            "payload_sample",
            "technical_information",
            "local_id",
            "title",
        ]
        return self.format_template(
            report,
            template=template,
            keys=[*additional_keys, *keys],
            html_parser=html_parser,
        )

    def set_yeswehack_domain(self, ywh_domain):
        self._ywh_domain = ywh_domain

    ############################################################
    ####################### Static methods #####################
    ############################################################
    @staticmethod
    def format_template(
        report, template, keys=[], fmt_keys={}, html_parser=html_handler
    ):
        """
        Format given template with report information

        :param yeswehack.api.Report report: Report representation
        :param str template: template to use.
        :param list additional_keys: keys corresponding to attribute of report object.

        to used subjobject, use '__' separator : cvss__score -> cvss.score
        """

        for key in keys:
            obj = report
            for i in key.split("__"):
                if obj:
                    obj = getattr(obj, i, None)
                else:
                    break
            if "html" in key:
                obj = html_parser(obj)
            fmt_keys[key] = obj
        return template.format(**fmt_keys)

    ############################################################
    ###################### Abstract methods ####################
    ############################################################
    @abstractmethod
    def configure(bugtracker):
        pass

    @abstractmethod
    def get_project(self):
        pass

    @abstractmethod
    def post_issue(self, report):
        pass

    @abstractmethod
    def get_url(self, issue):
        pass

    @abstractmethod
    def get_id(self, issue):
        pass



"""
Specific Exceptions
"""


class BugTrackerNotFound(Exception):
    """
    Exceptions for non exitsing Bugtracker
    """

    pass


class MultipleBugTrackerMacth(Exception):
    """
    Exception for multiple Bugtracker having the same type identifier.
    """

    pass



from vstsclient.vstsclient import VstsClient
from vstsclient.models import JsonPatchDocument, JsonPatchOperation
from vstsclient.constants import SystemFields, MicrosoftFields
from colorama import Fore, Style
from tfs.ywhtfs import YWHTfs, YWHTfsConfig, NewVstsClient
from ywh2bt.utils import read_input


class YWHAzuredevops(YWHTfs):
    def __define_client__(self, *args, **kwargs):
        self.bt = NewVstsClient(
            args[0], personal_access_token=args[1], scheme=kwargs["scheme"]
        )


class YWHAzuredevopsConfig(YWHTfsConfig):
    # base_url = "dev.azure.com"
    bugtracker_type = "azuredevops"
    client = YWHAzuredevops
    optional_keys = {**YWHTfsConfig.optional_keys, "url": "dev.azure.com"}

    def configure_url(self):
        url = (
            read_input(
                Fore.BLUE
                + self.type.title()
                + " url (default: {}):".format(self.optional_keys["url"])
                + Style.RESET_ALL
            )
            or self.optional_keys["url"]
        )
        orga = read_input(
            Fore.BLUE
            + "organisation name if you don't have set it in url '{url}/<orga>':".format(
                url=url
            )
            + Style.RESET_ALL
        )

        # project_types = self.bt.get_workitem_types(self.project_name)
        self._url = "{url}/{orga}".format(url=url, orga=orga)



from vstsclient.vstsclient import VstsClient
from vstsclient.models import JsonPatchDocument, JsonPatchOperation
from vstsclient.constants import SystemFields, MicrosoftFields
from vstsclient._http import HTTPError
from colorama import Fore, Style
from ywh2bt.trackers.bugtracker import BugTracker
from ywh2bt.utils import read_input
from ywh2bt.config import BugTrackerConfig


class NewVstsClient(VstsClient):
    """Override VstsClient Base class for add Scheme support"""

    def __init__(self, *args, scheme="HTTP", **kwargs):
        super(NewVstsClient, self).__init__(*args, **kwargs)
        self._http_client.protocol = scheme


class YWHTfs(BugTracker):
    def __init__(
        self,
        url,
        project,
        token,
        assigned_to,
        collection="DefaultCollection",
        issuetype="Task",
    ):
        self.__define_client__(
            url,
            token,
            scheme=url.split("://")[0].upper() or "HTTP",
            collection=collection,
        )

        try:
            projects = self.bt.get_projects()
        except HTTPError:
            raise

        self.issuetype = Task
        self.assigned_to = assigned_to

        if project:
            self.project = project
            if self.project not in [project.name for project in projects]:
                raise Exception(
                    "Project Name does'nt exist in {url}".format(config["url"])
                )

    def __define_client__(self, url, pat, collection, scheme="HTTP"):
        self.bt = NewVstsClient(
            url,
            scheme=scheme,
            personal_access_token=pat,
            collection=collection,
        )

    def get_project(self):
        try:
            repo = self.bt.get_project(self.project)
        except HTTPError:
            raise
        return repo

    def post_issue(self, report):
        description = self.description_template
        title = self.issue_name_template.format(
            report_local_id=report.local_id, report_title=report.title
        )
        body = description.format(
            end_point=report.end_point,
            vulnerable_part=report.vulnerable_part,
            cvss=report.cvss.score,
            bug_type=report.bug_type.category.name,
            bug_description=report.bug_type.description,
            remediation_link=report.bug_type.link,
            description=report.description_html,
        )
        doc = JsonPatchDocument()
        doc.add(JsonPatchOperation("add", SystemFields.TITLE, title))
        doc.add(JsonPatchOperation("add", SystemFields.DESCRIPTION, body))
        doc.add(JsonPatchOperation("add", SystemFields.TAGS, "ywh-report"))
        doc.add(
            JsonPatchOperation(
                "add", SystemFields.ASSIGNED_TO, self.assigned_to
            )
        )
        # Create a new work item by specifying the project and work item type
        try:
            workitem = self.bt.create_workitem(
                self.project, self.issuetype, doc
            )
        except HTTPError:
            raise
        return workitem

    def get_url(self, issue):
        return issue.url

    def get_id(self, issue):
        return issue.id


class YWHTfsConfig(BugTrackerConfig):

    bugtracker_type = "tfs"
    client = YWHTfs
    mandatory_keys = ["url", "project", "assigned_to"]
    secret_keys = ["token"]
    optional_keys = dict(issuetype="Task", collection="DefaultCollection")

    def user_config(self):
        if not self._url.endswith("/tfs"):
            self._url += "/tfs"

    def _set_bugtracker(self):
        self._get_bugtracker(
            self._url,
            self._project,
            self._token,
            self._assigned_to,
            issuetype=self._issuetype,
            collection=self._collection,
        )



# -*- encoding: utf-8 -*-

import gitlab
from .bugtracker import BugTracker
from ywh2bt.config import BugTrackerConfig
from copy import copy

__all__ = ["YWHGitlab", "YWHGitlabConfig"]


"""
GiLab Bugtracker System
"""


class YWHGitlab(BugTracker):

    """
    Gitlab Client Wrapper
    """

    ############################################################
    ####################### Constructor ########################
    ############################################################
    def __init__(self, url, project, token):
        self.url = url
        self.project = project
        self.token = token
        self.bt = gitlab.Gitlab(self.url, private_token=self.token)
        try:
            self.bt.auth()
        except gitlab.exceptions.GitlabAuthenticationError:
            raise

    ############################################################
    #################### Instance methods ######################
    ############################################################
    def get_project(self):
        try:
            project = self.bt.projects.get(self.project)
        except gitlab.exceptions.GitlabAuthenticationError:
            return False
        except gitlab.exceptions.GitlabGetError:
            return False
        return project

    def post_issue(self, report):
        copy_report = copy(report)
        project = self.bt.projects.get(self.project)
        for attachment in report.attachments:
            # Add attachment to gitlab issue if there is attachement in the original bug
            f = project.upload(attachment.original_name, attachment.data)
            copy_report.description_html = copy_report.description_html.replace(
                attachment.url, f["url"]
            )

        description = self.report_as_description(copy_report)
        issue_data = {
            "title": self.report_as_title(copy_report),
            "description": description,
        }
        issue = project.issues.create(issue_data)
        return issue

    def get_url(self, issue):
        return issue.web_url

    def get_id(self, issue):
        return issue.id


class YWHGitlabConfig(BugTrackerConfig):

    """
    BugTrackerConfig Class associated to gitlab

    :attr str bugtracker_type: bugtracker identifier
    :attr class client: Client class
    :attr list mandatory_keys: keys needed in configuration file part
    :attr list secret_keys: keys need interaction in interactive mode
    :attr dict optional_keys: keys with default values
    :attr dict _description: descriptor for each specified key
    """

    bugtracker_type = "gitlab"
    client = YWHGitlab

    mandatory_keys = ["project"]
    secret_keys = ["token"]
    optional_keys = dict(url="http://gitlab.com")
    _description = dict(project="path/to/project")

    ############################################################
    #################### Instance methods ######################
    ############################################################
    def _set_bugtracker(self):
        self._get_bugtracker(self._url, self._project, self._token)
